# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileNotice: Part of the AddonManager.

# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'compact_view.ui'
##
## Created by: Qt User Interface Compiler version 5.15.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySideWrapper import QtCore, QtWidgets


class Ui_CompactView(object):
    def setupUi(self, CompactView):
        if not CompactView.objectName():
            CompactView.setObjectName("CompactView")
        CompactView.resize(489, 16)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CompactView.sizePolicy().hasHeightForWidth())
        CompactView.setSizePolicy(sizePolicy)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(CompactView)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.horizontalLayout_2.setContentsMargins(3, 0, 9, 0)
        self.labelIcon = QtWidgets.QLabel(CompactView)
        self.labelIcon.setObjectName("labelIcon")
        sizePolicy1 = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed
        )
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.labelIcon.sizePolicy().hasHeightForWidth())
        self.labelIcon.setSizePolicy(sizePolicy1)
        self.labelIcon.setMinimumSize(QtCore.QSize(16, 16))
        self.labelIcon.setBaseSize(QtCore.QSize(16, 16))

        self.horizontalLayout_2.addWidget(self.labelIcon)

        self.labelPackageName = QtWidgets.QLabel(CompactView)
        self.labelPackageName.setObjectName("labelPackageName")

        self.labelPackageNameSpacer = QtWidgets.QLabel(CompactView)
        self.labelPackageNameSpacer.setText(" — ")
        self.labelPackageNameSpacer.setObjectName("labelPackageNameSpacer")

        self.horizontalLayout_2.addWidget(self.labelPackageName)
        self.horizontalLayout_2.addWidget(self.labelPackageNameSpacer)

        self.labelVersion = QtWidgets.QLabel(CompactView)
        self.labelVersion.setObjectName("labelVersion")

        self.horizontalLayout_2.addWidget(self.labelVersion)

        self.labelDescription = QtWidgets.QLabel(CompactView)
        self.labelDescription.setObjectName("labelDescription")
        sizePolicy2 = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.labelDescription.sizePolicy().hasHeightForWidth())
        self.labelDescription.setSizePolicy(sizePolicy2)
        self.labelDescription.setTextFormat(QtCore.Qt.PlainText)
        self.labelDescription.setWordWrap(False)
        self.labelDescription.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

        self.horizontalLayout_2.addWidget(self.labelDescription)

        self.labelStatus = QtWidgets.QLabel(CompactView)
        self.labelStatus.setObjectName("labelStatus")

        self.horizontalLayout_2.addWidget(self.labelStatus)

        self.retranslateUi(CompactView)

        QtCore.QMetaObject.connectSlotsByName(CompactView)

    # setupUi

    def retranslateUi(self, CompactView):
        #        CompactView.setWindowTitle(QCoreApplication.translate("CompactView", "Form", None))
        self.labelIcon.setText(QtCore.QCoreApplication.translate("CompactView", "Icon", None))
        self.labelPackageName.setText(
            QtCore.QCoreApplication.translate("CompactView", "<b>Package Name</b>", None)
        )
        self.labelVersion.setText(QtCore.QCoreApplication.translate("CompactView", "Version", None))
        self.labelDescription.setText(
            QtCore.QCoreApplication.translate("CompactView", "Description", None)
        )
        self.labelStatus.setText(
            QtCore.QCoreApplication.translate("CompactView", "Update available", None)
        )

    # retranslateUi



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to split wires into separate wire objects."""
## @package split
# \ingroup draftfunctions
# \brief Provides functions to split wires into separate wire objects.

## \addtogroup draftfunctions
# @{
from draftmake import make_copy
from draftutils import utils


def split(wire, newPoint, edgeIndex):
    if utils.get_type(wire) != "Wire":
        return None
    if wire.Closed:
        return split_closed_wire(wire, edgeIndex)
    return split_open_wire(wire, newPoint, edgeIndex)


def split_closed_wire(wire, edgeIndex):
    wire.Closed = False
    new = make_copy.make_copy(wire)
    if edgeIndex == len(wire.Points):
        new.Points = [wire.Points[0], wire.Points[-1]]
    else:
        new.Points = [wire.Points[edgeIndex - 1], wire.Points[edgeIndex]]
        wire.Points = list(reversed(wire.Points[0:edgeIndex])) + list(
            reversed(wire.Points[edgeIndex:])
        )
    return new


splitClosedWire = split_closed_wire


def split_open_wire(wire, newPoint, edgeIndex):
    new = make_copy.make_copy(wire)
    wire1Points = []
    wire2Points = []
    for index, point in enumerate(wire.Points):
        if index == edgeIndex:
            edge = wire.getSubObject("Edge" + str(edgeIndex))
            newPoint = wire.Placement.inverse().multVec(edge.Curve.projectPoint(newPoint))
            wire1Points.append(newPoint)
            wire2Points.append(newPoint)
            wire2Points.append(point)
        elif index < edgeIndex:
            wire1Points.append(point)
        elif index > edgeIndex:
            wire2Points.append(point)
    wire.Points = wire1Points
    new.Points = wire2Points
    return new


splitOpenWire = split_open_wire

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create a fusion of two shapes."""
## @package fuse
# \ingroup draftfunctions
# \brief Provides functions to create a fusion of two shapes.

## \addtogroup draftfunctions
# @{
import FreeCAD as App
import draftutils.gui_utils as gui_utils

from draftmake.make_wire import Wire

if App.GuiUp:
    from draftviewproviders.view_wire import ViewProviderWire


def fuse(object1, object2):
    """fuse(oject1, object2)

    Returns an object made from the union of the 2 given objects.
    If the objects are coplanar, a special Draft Wire is used, otherwise we use
    a standard Part fuse.

    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    import Part
    import DraftGeomUtils

    # testing if we have holes:
    holes = False
    fshape = object1.Shape.fuse(object2.Shape)
    fshape = fshape.removeSplitter()
    for f in fshape.Faces:
        if len(f.Wires) > 1:
            holes = True
    if DraftGeomUtils.isCoplanar(object1.Shape.fuse(object2.Shape).Faces) and not holes:
        obj = App.ActiveDocument.addObject("Part::FeaturePython", "Fusion")
        Wire(obj)
        if App.GuiUp:
            ViewProviderWire(obj.ViewObject)
        obj.Base = object1
        obj.Tool = object2
    elif holes:
        # temporary hack, since Part::Fuse objects don't remove splitters
        obj = App.ActiveDocument.addObject("Part::Feature", "Fusion")
        obj.Shape = fshape
    else:
        obj = App.ActiveDocument.addObject("Part::Fuse", "Fusion")
        obj.Base = object1
        obj.Tool = object2
    if App.GuiUp:
        object1.ViewObject.Visibility = False
        object2.ViewObject.Visibility = False
        gui_utils.format_object(obj, object1)
        gui_utils.select(obj)

    return obj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Unit tests for the ArchComponent module

import Arch
import Draft
import Part
import FreeCAD as App
from bimtests import TestArchBase
from draftutils.messages import _msg

from math import pi, cos, sin, radians


class TestArchComponent(TestArchBase.TestArchBase):

    def testAdd(self):
        App.Console.PrintLog("Checking Arch Add...\n")
        l = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(2, 0, 0))
        w = Arch.makeWall(l, width=0.2, height=2)
        sb = Part.makeBox(1, 1, 1)
        b = App.ActiveDocument.addObject("Part::Feature", "Box")
        b.Shape = sb
        App.ActiveDocument.recompute()
        Arch.addComponents(b, w)
        App.ActiveDocument.recompute()
        r = w.Shape.Volume > 1.5
        self.assertTrue(r, "Arch Add failed")

    def testRemove(self):
        App.Console.PrintLog("Checking Arch Remove...\n")
        l = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(2, 0, 0))
        w = Arch.makeWall(l, width=0.2, height=2, align="Right")
        sb = Part.makeBox(1, 1, 1)
        b = App.ActiveDocument.addObject("Part::Feature", "Box")
        b.Shape = sb
        App.ActiveDocument.recompute()
        Arch.removeComponents(b, w)
        App.ActiveDocument.recompute()
        r = w.Shape.Volume < 0.75
        self.assertTrue(r, "Arch Remove failed")

    def testBsplineSlabAreas(self):
        """Test the HorizontalArea and VerticalArea properties of a Bspline-based slab.

        See https://github.com/FreeCAD/FreeCAD/issues/20989.
        """

        operation = "Checking Bspline slab area calculation..."
        self.printTestMessage(operation)

        doc = App.ActiveDocument

        # Parameters
        radius = 10000  # 10 meters in mm
        extrusionLength = 100  # 10 cm in mm
        numPoints = 50  # Number of points for B-spline
        startAngle = 0  # Start at 0 degrees (right)
        endAngle = 180  # End at 180 degrees (left)

        # Create points for semicircle
        points = []
        angleStep = (endAngle - startAngle) / (numPoints - 1)
        for i in range(numPoints):
            angleDeg = startAngle + i * angleStep
            angleRad = radians(angleDeg)
            x = radius * cos(angleRad)
            y = radius * sin(angleRad)
            points.append(App.Vector(x, y, 0))

        # Create Draft objects
        bspline = Draft.makeBSpline(points, closed=False)
        closingLine = Draft.makeLine(points[-1], points[0])
        doc.recompute()

        # Create sketch
        # We do this because Draft.make_wires does not support B-splines
        # and we need a closed wire for the slab
        sketch = Draft.makeSketch([bspline, closingLine], autoconstraints=True, delete=True)
        if sketch is None:
            self.fail("Sketch creation failed")
        sketch.recompute()

        # Create slab
        slab = Arch.makeStructure(sketch, length=extrusionLength, name="Slab")
        slab.recompute()

        # Calculate theoretical areas
        radiusMeters = radius / 1000
        heightMeters = extrusionLength / 1000
        theoreticalHorizontalArea = (pi * radiusMeters**2) / 2
        theoreticalVerticalArea = (pi * radiusMeters + 2 * radiusMeters) * heightMeters

        # Get actual areas
        actualHorizontalArea = slab.HorizontalArea.getValueAs("m^2").Value
        actualVerticalArea = slab.VerticalArea.getValueAs("m^2").Value

        # Optimally wrapped assertions
        self.assertAlmostEqual(
            actualHorizontalArea,
            theoreticalHorizontalArea,
            places=3,
            msg=(
                "Horizontal area > 0.1% tolerance | "
                f"Exp: {theoreticalHorizontalArea:.3f} m² | "
                f"Got: {actualHorizontalArea:.3f} m²"
            ),
        )

        self.assertAlmostEqual(
            actualVerticalArea,
            theoreticalVerticalArea,
            places=3,
            msg=(
                "Vertical area > 0.1% tolerance | "
                f"Exp: {theoreticalVerticalArea:.3f} m² | "
                f"Got: {actualVerticalArea:.3f} m²"
            ),
        )

    def testHouseSpaceAreas(self):
        """Test the HorizontalArea and VerticalArea properties of a house-like space.

        See https://github.com/FreeCAD/FreeCAD/issues/14687.
        """

        operation = "Checking house space area calculation..."
        self.printTestMessage(operation)

        doc = App.ActiveDocument

        # Dimensional parameters (all in mm)
        baseLength = 5000  # 5m along X-axis
        baseWidth = 5000  # 5m along Y-axis (extrusion depth)
        rectangleHeight = 2500  # 2.5m lower rectangular portion
        triangleHeight = 2500  # 2.5m upper triangular portion
        totalHeight = rectangleHeight + triangleHeight  # 5m total height

        # Create envelope profile points (XZ plane)
        points = [
            App.Vector(0, 0, 0),
            App.Vector(baseLength, 0, 0),
            App.Vector(baseLength, 0, rectangleHeight),
            App.Vector(baseLength / 2, 0, totalHeight),
            App.Vector(0, 0, rectangleHeight),
        ]

        # Create wire with automatic face creation
        wire = Draft.makeWire(points, closed=True, face=True)
        if not wire:
            self.fail(f"Wire creation failed with points: {points}\n")
        doc.recompute()

        # Extrude the wire
        extrudedObj = Draft.extrude(wire, App.Vector(0, baseWidth, 0), solid=True)
        if not extrudedObj:
            self.fail("Extrusion failed - no object created\n")
        extrudedObj.Label = "Extruded house"
        doc.recompute()

        # Create Arch Space from the extrusion
        space = Arch.makeSpace(extrudedObj)
        space.Label = "House space"
        doc.recompute()

        # Calculate theoretical areas
        # Horizontal area (only bottom face on XY plane)
        theoreticalHorizontalArea = (baseLength * baseWidth) / 1e6  # 25 m²

        # Vertical areas
        # Side faces (YZ plane) - two rectangles
        sideFaceArea = (rectangleHeight * baseWidth) / 1e6  # 12.5 m² each
        totalSides = sideFaceArea * 2  # 25 m²

        # Front/back faces (XZ plane)
        rectangularPart = (baseLength * rectangleHeight) / 1e6  # 12.5 m²
        triangularPart = (baseLength * triangleHeight / 2) / 1e6  # 6.25 m²
        totalFrontBack = (rectangularPart + triangularPart) * 2  # 37.5 m²

        theoreticalVerticalArea = totalSides + totalFrontBack  # 62.5 m²

        # Get actual areas from space
        actualHorizontalArea = space.HorizontalArea.getValueAs("m^2").Value
        actualVerticalArea = space.VerticalArea.getValueAs("m^2").Value

        self.assertAlmostEqual(
            actualHorizontalArea,
            theoreticalHorizontalArea,
            places=3,
            msg=f"Horizontal area > 0.1% | Exp: {theoreticalHorizontalArea:.3f} | "
            f"Got: {actualHorizontalArea:.3f}",
        )

        self.assertAlmostEqual(
            actualVerticalArea,
            theoreticalVerticalArea,
            places=3,
            msg=f"Vertical area > 0.1% | Exp: {theoreticalVerticalArea:.3f} | "
            f"Got: {actualVerticalArea:.3f}",
        )

    def test_remove_single_window_from_wall_host_is_none(self):
        """
        Tests that a window is removed from its wall's host list when
        Arch.removeComponents is called with host=None (single window selection scenario).

        See https://github.com/FreeCAD/FreeCAD/issues/21551
        """

        # Create a basic wall
        wall_base = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(3000, 0, 0))
        wall = Arch.makeWall(wall_base, width=200, height=2500)

        # Create a window to be added to the wall
        window_width = 1000.0
        window_height = 1200.0

        # Create a Draft Rectangle for the window's Base. Arch.makeWindow can work with
        # either a Wire or a Face.
        window_base = Draft.makeRectangle(length=window_width, height=window_height)

        window = Arch.makeWindow(baseobj=window_base)
        window.Width = window_width  # Manually set as makeWindow(base) doesn't
        window.Height = window_height  # Manually set

        self.document.recompute()

        # Add the window to the wall.
        Arch.addComponents(window, wall)
        self.document.recompute()

        # Pre-condition check: ensure the wall is indeed a host for the window.
        self.assertIn(wall, window.Hosts, "Wall should be in window.Hosts before removal.")
        self.assertEqual(len(window.Hosts), 1, "Window should have 1 host before removal.")

        # Simulate the Arch_Remove command with the scenario where only the window
        # is selected and "Remove" is used.
        Arch.removeComponents([window], host=None)
        self.document.recompute()  # Important for Arch objects to update their state.

        # Assert: the wall should no longer be in the window's Hosts list.
        self.assertNotIn(wall, window.Hosts, "Wall should not be in window.Hosts after removal.")
        self.assertEqual(len(window.Hosts), 0, "Window.Hosts list should be empty after removal.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides general functions to work with topological shapes."""
## @package general
# \ingroup draftgeoutils
# \brief Provides general functions to work with topological shapes.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils
from draftutils import params

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{

# Default normal direction for all geometry operations
NORM = App.Vector(0, 0, 1)


def precision():
    """Return the Draft precision setting."""
    # Set precision level with a cap to avoid overspecification that:
    #  1 - whilst it is precise enough (e.g. that OCC would consider
    #      2 points are coincident)
    #      (not sure what it should be 10 or otherwise);
    #  2 - but FreeCAD/OCC can handle 'internally'
    #      (e.g. otherwise user may set something like
    #      15 that the code would never consider 2 points are coincident
    #      as internal float is not that precise)
    precisionMax = 10
    precisionInt = params.get_param("precision")
    precisionInt = precisionInt if precisionInt <= 10 else precisionMax
    return precisionInt


def vec(edge, use_orientation=False):
    """Return a vector from an edge or a Part.LineSegment.

    If use_orientation is True, it takes into account the edges orientation.
    If edge is not straight, you'll get strange results!
    """
    if isinstance(edge, Part.Shape):
        if use_orientation and isinstance(edge, Part.Edge) and edge.Orientation == "Reversed":
            return edge.Vertexes[0].Point.sub(edge.Vertexes[-1].Point)
        else:
            return edge.Vertexes[-1].Point.sub(edge.Vertexes[0].Point)
    elif isinstance(edge, Part.LineSegment):
        return edge.EndPoint.sub(edge.StartPoint)
    else:
        return None


def edg(p1, p2):
    """Return an edge from 2 vectors."""
    if isinstance(p1, App.Vector) and isinstance(p2, App.Vector):
        if DraftVecUtils.equals(p1, p2):
            return None

        return Part.LineSegment(p1, p2).toShape()


def getVerts(shape):
    """Return a list containing vectors of each vertex of the shape."""
    if not hasattr(shape, "Vertexes"):
        return []

    p = []
    for v in shape.Vertexes:
        p.append(v.Point)
    return p


def v1(edge):
    """Return the first point of an edge."""
    return edge.Vertexes[0].Point


def isNull(something):
    """Return True if the given shape, vector, or placement is Null.

    If the vector is (0, 0, 0), it will return True.
    """
    if isinstance(something, Part.Shape):
        return something.isNull()
    elif isinstance(something, App.Vector):
        if something == App.Vector(0, 0, 0):
            return True
        else:
            return False
    elif isinstance(something, App.Placement):
        if something.Base == App.Vector(0, 0, 0) and something.Rotation.Q == (0, 0, 0, 1):
            return True
        else:
            return False


def isPtOnEdge(pt, edge):
    """Test if a point lies on an edge."""
    v = Part.Vertex(pt)
    try:
        d = v.distToShape(edge)
    except Part.OCCError:
        return False
    else:
        if d:
            if round(d[0], precision()) == 0:
                return True
    return False


def hasCurves(shape):
    """Check if the given shape has curves."""
    for e in shape.Edges:
        if not isinstance(e.Curve, (Part.LineSegment, Part.Line)):
            return True
    return False


def isAligned(edge, axis="x"):
    """Check if the given edge or line is aligned to the given axis.

    The axis can be 'x', 'y' or 'z'.
    """

    def is_same(a, b):
        return round(a, precision()) == round(b, precision())

    if axis == "x":
        if isinstance(edge, Part.Edge):
            if len(edge.Vertexes) == 2:
                return is_same(edge.Vertexes[0].X, edge.Vertexes[-1].X)
        elif isinstance(edge, Part.LineSegment):
            return is_same(edge.StartPoint.x, edge.EndPoint.x)

    elif axis == "y":
        if isinstance(edge, Part.Edge):
            if len(edge.Vertexes) == 2:
                return is_same(edge.Vertexes[0].Y, edge.Vertexes[-1].Y)
        elif isinstance(edge, Part.LineSegment):
            return is_same(edge.StartPoint.y, edge.EndPoint.y)

    elif axis == "z":
        if isinstance(edge, Part.Edge):
            if len(edge.Vertexes) == 2:
                return is_same(edge.Vertexes[0].Z, edge.Vertexes[-1].Z)
        elif isinstance(edge, Part.LineSegment):
            return is_same(edge.StartPoint.z, edge.EndPoint.z)

    return False


def getQuad(face):
    """Return a list of 3 vectors if the face is a quad, otherwise None.

    Returns
    -------
    basepoint, Xdir, Ydir
        If the face is a quad.

    None
        If the face is not a quad.
    """
    if len(face.Edges) != 4:
        return None

    v1 = vec(face.Edges[0])  # Warning redefinition of function v1
    v2 = vec(face.Edges[1])
    v3 = vec(face.Edges[2])
    v4 = vec(face.Edges[3])
    angles90 = [round(math.pi * 0.5, precision()), round(math.pi * 1.5, precision())]

    angles180 = [0, round(math.pi, precision()), round(math.pi * 2, precision())]
    for ov in [v2, v3, v4]:
        if not (round(v1.getAngle(ov), precision()) in angles90 + angles180):
            return None

    for ov in [v2, v3, v4]:
        if round(v1.getAngle(ov), precision()) in angles90:
            v1.normalize()
            ov.normalize()
            return [face.Edges[0].Vertexes[0].Point, v1, ov]


def areColinear(e1, e2):
    """Return True if both edges are colinear."""
    if not isinstance(e1.Curve, (Part.LineSegment, Part.Line)):
        return False
    if not isinstance(e2.Curve, (Part.LineSegment, Part.Line)):
        return False

    v1 = vec(e1)
    v2 = vec(e2)
    a = round(v1.getAngle(v2), precision())
    if (a == 0) or (a == round(math.pi, precision())):
        v3 = e2.Vertexes[0].Point.sub(e1.Vertexes[0].Point)
        if DraftVecUtils.isNull(v3):
            return True
        else:
            a2 = round(v1.getAngle(v3), precision())
            if (a2 == 0) or (a2 == round(math.pi, precision())):
                return True
    return False


def hasOnlyWires(shape):
    """Return True if all edges are inside a wire."""
    ne = 0
    for w in shape.Wires:
        ne += len(w.Edges)
    if ne == len(shape.Edges):
        return True
    return False


def geomType(edge):
    """Return the type of geometry this edge is based on.

    Parameters
    ----------
    edge: the edge whose `Curve` attribute is to be checked.

    Returns
    -------
    str
        Return the type of the edge's Curve attribute or "Unknown",
        if the parameter is missing.
    """
    try:
        if isinstance(edge.Curve, (Part.LineSegment, Part.Line)):
            return "Line"
        elif isinstance(edge.Curve, Part.Circle):
            return "Circle"
        elif isinstance(edge.Curve, Part.BSplineCurve):
            return "BSplineCurve"
        elif isinstance(edge.Curve, Part.BezierCurve):
            return "BezierCurve"
        elif isinstance(edge.Curve, Part.Ellipse):
            return "Ellipse"
        else:
            return "Unknown"
    except Exception:  # catch all errors, no only TypeError
        return "Unknown"


def isValidPath(shape):
    """Return True if the shape can be used as an extrusion path."""
    if shape.isNull():
        return False
    if shape.Faces:
        return False
    if len(shape.Wires) > 1:
        return False
    if shape.Wires:
        if shape.Wires[0].isClosed():
            return False
    if shape.isClosed():
        return False
    return True


def findClosest(base_point, point_list):
    """Find closest point in a list of points to the base point.

    Returns
    -------
    int
        An index from the list of points is returned.

    None
        If point_list is empty.
    """
    npoint = None
    if not point_list:
        return None

    smallest = 1000000
    for n in range(len(point_list)):
        new = base_point.sub(point_list[n]).Length
        if new < smallest:
            smallest = new
            npoint = n

    return npoint


def getBoundaryAngles(angle, alist):
    """Return the 2 closest angles that encompass the given angle."""
    negs = True
    while negs:
        negs = False
        for i in range(len(alist)):
            if alist[i] < 0:
                alist[i] = 2 * math.pi + alist[i]
                negs = True
        if angle < 0:
            angle = 2 * math.pi + angle
            negs = True

    lower = None
    for a in alist:
        if a < angle:
            if lower is None:
                lower = a
            else:
                if a > lower:
                    lower = a

    if lower is None:
        lower = 0
        for a in alist:
            if a > lower:
                lower = a

    higher = None
    for a in alist:
        if a > angle:
            if higher is None:
                higher = a
            else:
                if a < higher:
                    higher = a

    if higher is None:
        higher = 2 * math.pi
        for a in alist:
            if a < higher:
                higher = a

    return lower, higher


## @}



from typing import Any, Final

from Base.Metadata import export

from Gui.ViewProviderGeometryObject import ViewProviderGeometryObject

@export(
    Include="Mod/Fem/Gui/ViewProviderFemConstraint.h",
    Namespace="FemGui",
)
class ViewProviderFemConstraint(ViewProviderGeometryObject):
    """
    This is the ViewProviderFemConstraint class

    Author: Mario Passaglia (mpassaglia@cbc.uba.ar)
    License: LGPL-2.1-or-later
    """

    def loadSymbol(self) -> Any:
        """loadSymbol(filename) -> None

        Load constraint symbol from Open Inventor file.
        The file structure should be as follows:
        A separator containing a separator with the symbol used in
        multiple copies at points on the surface and an optional
        separator with a symbol excluded from multiple copies.

        filename : str
            Open Inventor file."""
        ...
    SymbolNode: Final[Any]
    """A pivy SoSeparator with the nodes of the constraint symbols"""

    ExtraSymbolNode: Final[Any]
    """A pivy SoSeparator with the nodes of the constraint extra symbols"""

    RotateSymbol: bool
    """Apply rotation on copies of the constraint symbol"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Profile as PathProfile
import PathGui
from PySide.QtCore import QT_TRANSLATE_NOOP


__title__ = "CAM Profile Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Profile operation page controller and command implementation."


FeatureSide = 0x01
FeatureProcessing = 0x02


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Base class for profile operation page controllers. Two sub features are supported:
    FeatureSide       ... Is the Side property exposed in the UI
    FeatureProcessing ... Are the processing check boxes supported by the operation
    """

    # def initPage(self, obj):
    #     self.setTitle("Profile - " + obj.Label)
    #     self.updateVisibility()

    def profileFeatures(self):
        """profileFeatures() ... return which of the optional profile features are supported.
        Currently two features are supported and returned:
            FeatureSide       ... Is the Side property exposed in the UI
            FeatureProcessing ... Are the processing check boxes supported by the operation
        ."""
        return FeatureSide | FeatureProcessing

    def getForm(self):
        """getForm() ... returns UI customized according to profileFeatures()"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpProfileFullEdit.ui")

        comboToPropertyMap = [("cutSide", "Side"), ("direction", "Direction")]
        enumTups = PathProfile.ObjectProfile.areaOpPropertyEnumerations(dataType="raw")

        self.populateCombobox(form, enumTups, comboToPropertyMap)
        return form

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

        if obj.Side != str(self.form.cutSide.currentData()):
            obj.Side = str(self.form.cutSide.currentData())
        if obj.Direction != str(self.form.direction.currentData()):
            obj.Direction = str(self.form.direction.currentData())
        PathGuiUtil.updateInputField(obj, "OffsetExtra", self.form.extraOffset)
        obj.NumPasses = self.form.numPasses.value()
        PathGuiUtil.updateInputField(obj, "Stepover", self.form.stepover)

        if obj.UseComp != self.form.useCompensation.isChecked():
            obj.UseComp = self.form.useCompensation.isChecked()
        if obj.UseStartPoint != self.form.useStartPoint.isChecked():
            obj.UseStartPoint = self.form.useStartPoint.isChecked()

        if obj.processHoles != self.form.processHoles.isChecked():
            obj.processHoles = self.form.processHoles.isChecked()
        if obj.processPerimeter != self.form.processPerimeter.isChecked():
            obj.processPerimeter = self.form.processPerimeter.isChecked()
        if obj.processCircles != self.form.processCircles.isChecked():
            obj.processCircles = self.form.processCircles.isChecked()

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)

        self.selectInComboBox(obj.Side, self.form.cutSide)
        self.selectInComboBox(obj.Direction, self.form.direction)
        self.form.extraOffset.setText(
            FreeCAD.Units.Quantity(obj.OffsetExtra.Value, FreeCAD.Units.Length).UserString
        )
        self.form.numPasses.setValue(obj.NumPasses)
        self.form.stepover.setText(
            FreeCAD.Units.Quantity(obj.Stepover.Value, FreeCAD.Units.Length).UserString
        )

        self.form.useCompensation.setChecked(obj.UseComp)
        self.form.useStartPoint.setChecked(obj.UseStartPoint)
        self.form.processHoles.setChecked(obj.processHoles)
        self.form.processPerimeter.setChecked(obj.processPerimeter)
        self.form.processCircles.setChecked(obj.processCircles)

        self.updateVisibility()

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        signals.append(self.form.cutSide.currentIndexChanged)
        signals.append(self.form.direction.currentIndexChanged)
        signals.append(self.form.extraOffset.editingFinished)
        signals.append(self.form.numPasses.editingFinished)
        signals.append(self.form.stepover.editingFinished)
        if hasattr(self.form.useCompensation, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.useCompensation.checkStateChanged)
            signals.append(self.form.useStartPoint.checkStateChanged)
            signals.append(self.form.processHoles.checkStateChanged)
            signals.append(self.form.processPerimeter.checkStateChanged)
            signals.append(self.form.processCircles.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.useCompensation.stateChanged)
            signals.append(self.form.useStartPoint.stateChanged)
            signals.append(self.form.processHoles.stateChanged)
            signals.append(self.form.processPerimeter.stateChanged)
            signals.append(self.form.processCircles.stateChanged)

        return signals

    def updateVisibility(self):
        hasFace = False
        objBase = list()

        if hasattr(self.obj, "Base"):
            objBase = self.obj.Base

        if objBase.__len__() > 0:
            for base, subsList in objBase:
                for sub in subsList:
                    if sub[:4] == "Face":
                        hasFace = True
                        break

        if hasFace:
            self.form.processCircles.show()
            self.form.processHoles.show()
            self.form.processPerimeter.show()
        else:
            self.form.processCircles.hide()
            self.form.processHoles.hide()
            self.form.processPerimeter.hide()

        self.form.stepover.setEnabled(self.obj.NumPasses > 1)

    def registerSignalHandlers(self, obj):
        if hasattr(self.form.useCompensation, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.useCompensation.checkStateChanged.connect(self.updateVisibility)
        else:  # Qt version < 6.7.0
            self.form.useCompensation.stateChanged.connect(self.updateVisibility)
        self.form.numPasses.editingFinished.connect(self.updateVisibility)


# Eclass


Command = PathOpGui.SetupOperation(
    "Profile",
    PathProfile.Create,
    TaskPanelOpPage,
    "CAM_Profile",
    QT_TRANSLATE_NOOP("CAM_Profile", "Profile"),
    QT_TRANSLATE_NOOP("CAM_Profile", "Profile entire model, selected face(s) or selected edge(s)"),
    PathProfile.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathProfileGui ... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""The Addon Catalog is the main list of all Addons along with their various
sources and compatible versions. Added in FreeCAD 1.1 to replace .gitmodules."""

import base64
import datetime
import os
import xml.etree.ElementTree
from dataclasses import dataclass
import json
from hashlib import sha256
from typing import Any, Dict, List, Optional

from addonmanager_metadata import Version, MetadataReader
from Addon import Addon

import addonmanager_freecad_interface as fci


INVALID_TIME = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc).astimezone()


@dataclass
class CatalogEntryMetadata:
    """All contents of the metadata are the text contents of the file listed. The icon data is
    base64-encoded (even though it was probably an SVG, technically other formats are supported)."""

    package_xml: str = ""
    requirements_txt: str = ""
    metadata_txt: str = ""
    icon_data: str = ""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "CatalogEntryMetadata":
        """Create CatalogEntryMetadata from a data dictionary"""
        md = CatalogEntryMetadata()
        if "package_xml" in data:
            md.package_xml = data["package_xml"]
        if "requirements_txt" in data:
            md.requirements_txt = data["requirements_txt"]
        if "metadata_txt" in data:
            md.metadata_txt = data["metadata_txt"]
        if "icon_data" in data:
            md.icon_data = data["icon_data"]
        return md


@dataclass
class AddonCatalogEntry:
    """Each entry in the catalog, storing data about a particular version of an
    Addon. Note that this class needs to be identical to the one that is used in the remote cache
    generation, so don't make changes here without ensuring that the classes are synchronized."""

    freecad_min: Optional[Version] = None
    freecad_max: Optional[Version] = None
    repository: Optional[str] = None
    git_ref: Optional[str] = None
    zip_url: Optional[str] = None
    note: Optional[str] = None
    branch_display_name: Optional[str] = None
    metadata: Optional[CatalogEntryMetadata] = None  # Generated by the cache system
    last_update_time: str = ""  # Generated by the cache system
    curated: bool = True  # Generated by the cache system
    sparse_cache: bool = False  # Generated by the cache system
    relative_cache_path: str = ""  # Generated by the cache system

    def __init__(self, raw_data: Dict[str, str]) -> None:
        """Create an AddonDictionaryEntry from the raw JSON data"""
        super().__init__()
        for key, value in raw_data.items():
            if hasattr(self, key):
                if key in ("freecad_min", "freecad_max"):
                    if isinstance(value, str):
                        value = Version(from_string=value)
                    elif isinstance(value, list):
                        value = Version(from_list=value)
                    elif value is None:
                        pass
                    elif isinstance(value, dict) and "version_as_list" in value:
                        value = Version(from_list=value["version_as_list"])
                    else:
                        raise ValueError(f"Invalid value for {key}: {value}")
                elif key == "metadata":
                    if isinstance(value, dict):
                        metadata = CatalogEntryMetadata()
                        metadata.__dict__.update(value)
                        value = metadata
                    elif isinstance(value, str):
                        value = CatalogEntryMetadata.from_dict(json.loads(value))
                elif key == "git_ref" and self.branch_display_name is None:
                    self.branch_display_name = value
                setattr(self, key, value)

    def is_compatible(self) -> bool:
        """Check whether this AddonCatalogEntry is compatible with the current version of FreeCAD"""
        if self.freecad_min is None and self.freecad_max is None:
            return True
        current_version = Version(from_list=fci.Version())
        if self.freecad_min is None:
            return current_version <= self.freecad_max
        if self.freecad_max is None:
            return current_version >= self.freecad_min
        return self.freecad_min <= current_version <= self.freecad_max

    def unique_identifier(self) -> str:
        """Return a unique identifier of the AddonCatalogEntry, guaranteed to be repeatable: when
        given the same basic information, the same ID is created. Used as the key when storing
        the metadata for a given AddonCatalogEntry."""
        sha256_hash = sha256()
        sha256_hash.update(str(self).encode("utf-8"))
        return sha256_hash.hexdigest()

    def instantiate_addon(self, addon_id: str) -> Addon:
        """Return an instantiated Addon object"""
        if AddonCatalogEntry.is_installed(addon_id):
            state = Addon.Status.UNCHECKED
        else:
            state = Addon.Status.NOT_INSTALLED
        if self.sparse_cache:
            if self.zip_url:
                url = self.zip_url
            else:
                # Technically, this should never happen, but just in case...
                raise RuntimeError(f"Sparse cache entry {addon_id} has no zip_url")
        elif self.repository:
            url = self.repository
        else:
            url = self.zip_url
        if self.git_ref:
            addon = Addon(addon_id, url, state, branch=self.git_ref)
        else:
            addon = Addon(addon_id, url, state)
        addon.relative_cache_path = self.relative_cache_path

        if self.metadata:
            try:
                self._load_addon_metadata(addon, self.metadata)
            except xml.etree.ElementTree.ParseError:
                fci.Console.PrintWarning(
                    "An invalid or corrupted package.xml file was installed "
                    f"for {addon.display_name}\n"
                )

        try:
            addon.remote_last_updated = datetime.datetime.fromisoformat(self.last_update_time)
        except ValueError:
            addon.remote_last_updated = INVALID_TIME
        if state == Addon.Status.UNCHECKED:
            try:
                package_file = os.path.join(fci.DataPaths().mod_dir, addon_id, "package.xml")
                addon.installed_metadata = MetadataReader.from_file(package_file)
            except (FileNotFoundError, xml.etree.ElementTree.ParseError, RuntimeError):
                pass  # If there was an error, just ignore it, no metadata is not fatal

            most_recent_mtime = AddonCatalogEntry.most_recent_mtime(addon_id)
            addon.updated_timestamp = most_recent_mtime.timestamp()
            if most_recent_mtime < addon.remote_last_updated:
                addon.set_status(Addon.Status.UPDATE_AVAILABLE)
            elif (
                addon.installed_metadata
                and addon.metadata
                and addon.installed_metadata.version < addon.metadata.version
            ):
                # Fallback: even if the modification time is newer, if the stated version is older
                # something strange has happened, but flag an available update
                addon.set_status(Addon.Status.UPDATE_AVAILABLE)
            else:
                addon.set_status(Addon.Status.NO_UPDATE_AVAILABLE)

        addon.branch_display_name = (
            self.branch_display_name if self.branch_display_name else self.git_ref
        )

        addon.curated = self.curated

        return addon

    @staticmethod
    def is_installed(addon_id: str) -> bool:
        """Check if the addon is installed"""
        addon_dir = os.path.join(fci.DataPaths().mod_dir, addon_id)
        return os.path.exists(addon_dir) and os.listdir(addon_dir)

    @staticmethod
    def most_recent_mtime(addon_id: str) -> datetime.datetime:
        """Get the last update time of the addon by checking the modification time of all its
        files and returning the latest one."""
        if not AddonCatalogEntry.is_installed(addon_id):
            return INVALID_TIME
        addon_dir = os.path.join(fci.DataPaths().mod_dir, addon_id)
        max_time = INVALID_TIME
        for root, dirs, files in os.walk(addon_dir):
            dirs[:] = [d for d in dirs if d != ".git"]
            for file in files:
                file_path = os.path.join(root, file)
                try:
                    file_timestamp = os.path.getmtime(file_path)
                except OSError:
                    continue
                file_time = datetime.datetime.fromtimestamp(file_timestamp).astimezone()
                if file_time > max_time:
                    max_time = file_time
        return max_time

    def _load_addon_metadata(self, addon: Addon, cem: CatalogEntryMetadata):
        if cem.package_xml:
            metadata = MetadataReader.from_bytes(cem.package_xml.encode("utf-8"))
            addon.set_metadata(metadata)
        if cem.requirements_txt:
            AddonCatalogEntry._load_requirements_txt(addon, cem.requirements_txt)
        if cem.metadata_txt:
            AddonCatalogEntry._load_metadata_txt(addon, cem.metadata_txt)
        if cem.icon_data:
            self._load_icon_data(addon, cem.icon_data)

    @staticmethod
    def _load_metadata_txt(repo: Addon, data: str):
        """Process the metadata.txt metadata file"""
        lines = data.splitlines()
        for line in lines:
            if line.startswith("workbenches="):
                workbench_dependencies = line.split("=")[1].split(",")
                for wb in workbench_dependencies:
                    wb_name = wb.strip()
                    if wb_name:
                        repo.requires.add(wb_name)

            elif line.startswith("pylibs="):
                python_dependencies = line.split("=")[1].split(",")
                for pl in python_dependencies:
                    dep = pl.strip()
                    if dep:
                        repo.python_requires.add(dep)

            elif line.startswith("optionalpylibs="):
                optional_python_dependencies = line.split("=")[1].split(",")
                for pl in optional_python_dependencies:
                    dep = pl.strip()
                    if dep:
                        repo.python_optional.add(dep)

    @staticmethod
    def _load_requirements_txt(repo: Addon, data: str):
        """Process the requirements.txt metadata file"""

        lines = data.splitlines()
        for line in lines:
            break_chars = " <>=~!+#"
            package = line
            for n, c in enumerate(line):
                if c in break_chars:
                    package = line[:n].strip()
                    break
            if package:
                repo.python_requires.add(package)

    @staticmethod
    def _load_icon_data(repo: Addon, data: str):
        """Process the icon data."""

        repo.icon_data = base64.b64decode(data)
        if not repo.icon_data:
            raise ValueError(f"Invalid icon data '{data}' in cache for addon '{repo.name}'")


class AddonCatalog:
    """A catalog of addons grouped together into sets representing versions that are
    compatible with different versions of FreeCAD and/or represent different available branches
    of a given addon (e.g., a Development branch that users are presented)."""

    def __init__(self, data: Dict[str, Any]):
        self._original_data = data
        self._dictionary: Dict[str, List[AddonCatalogEntry]] = {}
        self._parse_raw_data()
        self._temp_icon_files = []

    def _parse_raw_data(self):
        self._dictionary = {}  # Clear pre-existing contents
        for key, value in self._original_data.items():
            if key in ["_meta", "$schema"]:  # Don't add the documentation objects to the tree
                continue
            self._dictionary[key] = []
            for entry in value:
                try:
                    self._dictionary[key].append(AddonCatalogEntry(entry))
                except (RuntimeError, LookupError) as e:
                    fci.Console.PrintWarning(
                        f"Failed to parse AddonCatalogEntry for {key}:\n\n {entry} \n\n"
                    )
                    fci.Console.PrintWarning(f"{e}\n")

    def get_available_addon_ids(self) -> List[str]:
        """Get a list of IDs that have at least one entry compatible with the current version of
        FreeCAD"""
        id_list = []
        for key, value in self._dictionary.items():
            for entry in value:
                if entry.is_compatible():
                    id_list.append(key)
                    break
        return id_list

    def get_all_addon_ids(self) -> List[str]:
        """Get a list of all Addon IDs, even those that have no compatible versions for the current
        version of FreeCAD."""
        id_list = []
        for key, value in self._dictionary.items():
            if len(value) == 0:
                continue
            id_list.append(key)
        return id_list

    def add_metadata_to_entry(
        self, addon_id: str, index: int, metadata: CatalogEntryMetadata
    ) -> None:
        """Adds metadata to an AddonCatalogEntry"""
        if addon_id not in self._dictionary:
            raise RuntimeError(f"Addon {addon_id} does not exist")
        if index >= len(self._dictionary[addon_id]):
            raise RuntimeError(f"Addon {addon_id} index out of range")
        self._dictionary[addon_id][index].metadata = metadata

    def get_available_branches(self, addon_id: str) -> List[str]:
        """For a given ID, get the list of available branches compatible with this version of
        FreeCAD.
        :return: A list of branch display names (or git refs, if no display name is available)"""
        if addon_id not in self._dictionary:
            return []
        result = []
        for entry in self._dictionary[addon_id]:
            if entry.is_compatible():
                result.append(
                    entry.branch_display_name if entry.branch_display_name else entry.git_ref
                )
        return result

    def get_catalog(self) -> Dict[str, List[AddonCatalogEntry]]:
        """Get access to the entire catalog without any filtering applied."""
        return self._dictionary

    def get_addon_from_id(self, addon_id: str, branch_display_name: Optional[str] = None) -> Addon:
        """Get the instantiated Addon object for the given ID and optionally branch. If no
        branch is provided, whichever branch is the "primary" branch will be returned (i.e., the
        first branch that matches). Raises a ValueError if no addon matches the request."""
        if addon_id not in self._dictionary:
            raise ValueError(f"Addon '{addon_id}' not found")
        for entry in self._dictionary[addon_id]:
            if not entry.is_compatible():
                continue
            if not branch_display_name or entry.branch_display_name == branch_display_name:
                return entry.instantiate_addon(addon_id)
        raise ValueError(
            f"Addon '{addon_id}' has no compatible branches named '{branch_display_name}'"
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Slab command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Slab:

    def __init__(self):
        self.callback = None
        self.view = None

    def GetResources(self):
        return {
            "Pixmap": "BIM_Slab",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Slab", "Slab"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Slab", "Creates a slab from a planar shape"),
            "Accel": "S,B",
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import DraftTools

        self.removeCallback()
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            self.proceed()
        else:
            if hasattr(FreeCADGui, "draftToolBar"):
                FreeCADGui.draftToolBar.selectUi()
            FreeCAD.Console.PrintMessage(translate("BIM", "Select a planar object") + "\n")
            FreeCAD.activeDraftCommand = self
            self.view = FreeCADGui.ActiveDocument.ActiveView
            self.callback = self.view.addEventCallback("SoEvent", DraftTools.selectObject)

    def proceed(self):
        self.removeCallback()
        sel = FreeCADGui.Selection.getSelection()
        if len(sel) == 1:
            FreeCADGui.addModule("Arch")
            FreeCAD.ActiveDocument.openTransaction("Create Slab")
            FreeCADGui.doCommand(
                "s = Arch.makeStructure(FreeCAD.ActiveDocument." + sel[0].Name + ",height=200)"
            )
            FreeCADGui.doCommand('s.Label = "' + translate("BIM", "Slab") + '"')
            FreeCADGui.doCommand('s.IfcType = "Slab"')
            FreeCADGui.doCommand("s.Normal = FreeCAD.Vector(0,0,-1)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()
        self.finish()

    def removeCallback(self):
        if self.callback:
            try:
                self.view.removeEventCallback("SoEvent", self.callback)
            except RuntimeError:
                pass
            self.callback = None

    def finish(self):
        self.removeCallback()
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.offUi()


FreeCADGui.addCommand("BIM_Slab", BIM_Slab())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import Path.Op.Base as PathOp
import Path.Op.EngraveBase as PathEngraveBase
import PathScripts.PathUtils as PathUtils
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = "Class and implementation of CAM Vcarve operation"

PRIMARY = 0
SECONDARY = 1
EXTERIOR1 = 2
EXTERIOR2 = 3
COLINEAR = 4
TWIN = 5
BORDERLINE = 6

# There is a bug in logging library. To enable debugging - set True also in Gui/Vcarve.py

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


def _collectVoronoiWires(vd):
    edges = [e for e in vd.Edges if e.Color == PRIMARY]
    vertex = {}
    for e in edges:
        for v in e.Vertices:
            i = v.Index
            j = vertex.get(i, [])
            j.append(e)
            vertex[i] = j

    # knots are the start and end points of a wire
    knots = [i for i in vertex if len(vertex[i]) == 1]
    knots.extend([i for i in vertex if len(vertex[i]) > 2])
    if len(knots) == 0:
        for i in vertex:
            if len(vertex[i]) > 0:
                knots.append(i)
                break

    def consume(v, edge):
        vertex[v] = [e for e in vertex[v] if e.Index != edge.Index]
        return len(vertex[v]) == 0

    def traverse(vStart, edge, edges):
        if vStart == edge.Vertices[0].Index:
            vEnd = edge.Vertices[1].Index
            edges.append(edge)
        else:
            vEnd = edge.Vertices[0].Index
            edges.append(edge.Twin)

        consume(vStart, edge)
        if consume(vEnd, edge):
            return None
        return vEnd

    wires = []
    while knots:
        we = []
        vFirst = knots[0]
        vStart = vFirst
        vLast = vFirst
        if len(vertex[vStart]):
            while vStart is not None:
                vLast = vStart
                edges = vertex[vStart]
                if len(edges) > 0:
                    edge = edges[0]
                    vStart = traverse(vStart, edge, we)
                else:
                    vStart = None
            wires.append(we)
        if len(vertex[vFirst]) == 0:
            knots = [v for v in knots if v != vFirst]
        if len(vertex[vLast]) == 0:
            knots = [v for v in knots if v != vLast]
    return wires


def _sortVoronoiWires(wires, start=FreeCAD.Vector(0, 0, 0)):
    def closestTo(start, point):
        p = None
        length = None
        for i in point:
            if length is None or length > start.distanceToPoint(point[i]):
                length = start.distanceToPoint(point[i])
                p = i
        return (p, length)

    begin = {}
    end = {}

    for i, w in enumerate(wires):
        begin[i] = w[0].Vertices[0].toPoint()
        end[i] = w[-1].Vertices[1].toPoint()

    result = []
    while begin:
        (bIdx, bLen) = closestTo(start, begin)
        (eIdx, eLen) = closestTo(start, end)
        if bLen < eLen:
            result.append(wires[bIdx])
            start = end[bIdx]
            del begin[bIdx]
            del end[bIdx]
        else:
            result.append([e.Twin for e in reversed(wires[eIdx])])
            start = begin[eIdx]
            del begin[eIdx]
            del end[eIdx]

    return result


class _Geometry(object):
    """POD class so the limits only have to be calculated once."""

    def __init__(self, zStart, zStop, zScale, zStepDown):
        self.start = zStart
        self.stop = zStop
        self.scale = zScale
        self.stepDown = zStepDown
        self.stepDownPass = 1

        # offset is used in finishing passes to override
        # any calculated vcarving depths. Usually going deeper 0.1-0.2 mm on finishing pass can help
        # remove "fuzzy skin" or other imperfections.
        self.offset = 0

    def incrementStepDownDepth(self, maximumUsableDepth):
        """
        Increase stepDown depth before starting new carving pass.
        :returns: True if successful, False if maximum depth achieved
        """

        # do not allow one to increase depth if we are already at stop depth
        if self.maximumDepth == self.stop:
            return False

        # do not allow one to increase depth if we are already at
        # maximum usable depth

        if self.maximumDepth <= maximumUsableDepth:
            return False

        self.stepDownPass += 1
        return True

    @property
    def maximumDepth(self):
        """
        Return maximum vcarving depth computed from step down setting and pass number
        """

        if self.stepDown == 0:
            return self.stop

        return max(self.stop, self.start - (self.stepDownPass * self.stepDown))

    @classmethod
    def FromTool(cls, tool, zStart, zFinal, zStepDown=0):
        rMax = float(tool.Diameter) / 2.0
        rMin = float(tool.TipDiameter) / 2.0
        toolangle = math.tan(math.radians(tool.CuttingEdgeAngle.Value / 2.0))
        zScale = 1.0 / toolangle
        zStop = zStart - rMax * zScale
        zOff = rMin * zScale

        return _Geometry(zStart + zOff, max(zStop + zOff, zFinal), zScale, zStepDown)

    @classmethod
    def FromObj(cls, obj, model):
        if obj.BaseShapes and hasattr(obj.BaseShapes[0], "Shape"):
            zStart = obj.BaseShapes[0].Shape.BoundBox.ZMax
        elif obj.Base and obj.Base[0][0] and hasattr(obj.Base[0][0], "Shape"):
            if len(obj.Base[0]) > 1 and "Face" in obj.Base[0][1][0]:
                faceName = obj.Base[0][1][0]
                faceIndex = int(faceName.replace("Face", "")) - 1
                face = obj.Base[0][0].Shape.Faces[faceIndex]
                zStart = face.BoundBox.ZMax
            else:
                zStart = obj.Base[0][0].Shape.BoundBox.ZMax
        else:
            zStart = model.Shape.BoundBox.ZMax
            Path.Log.error("Base object not set")
        finalDepth = obj.FinalDepth.Value
        stepDown = abs(obj.StepDown.Value)

        return cls.FromTool(obj.ToolController.Tool, zStart, finalDepth, stepDown)


def _calculate_depth(MIC, geom):
    # given a maximum inscribed circle (MIC) and tool angle,
    # return depth of cut relative to zStart.
    depth = geom.start - round(MIC * geom.scale, 4)

    return max(depth, geom.maximumDepth) + geom.offset


def _get_maximumUsableDepth(wires, geom):
    """
    Calculate maximum engraving depth for a list of wires
    belonging to one face.
    """

    def _get_depth(MIC, geom):
        """Similar logic to _calculate_depth but without stepdown and offset calculations"""
        depth = geom.start - round(MIC * geom.scale, 4)
        return max(depth, geom.stop)

    min_depth = None

    for wire in wires:
        for edge in wire:
            dist = edge.getDistances()
            depth = min(_get_depth(dist[0], geom), _get_depth(dist[1], geom))

            if min_depth is None:
                min_depth = depth
            else:
                min_depth = min(min_depth, depth)

    return min_depth


def _getPartEdge(edge, geom):
    dist = edge.getDistances()
    zBegin = _calculate_depth(dist[0], geom)
    zEnd = _calculate_depth(dist[1], geom)
    return edge.toShape(zBegin, zEnd)


def _getPartEdges(obj, vWire, geom):
    edges = []
    for e in vWire:
        edges.append(_getPartEdge(e, geom))
    return edges


class ObjectVcarve(PathEngraveBase.ObjectOp):
    """Proxy class for Vcarve operation."""

    def opFeatures(self, obj):
        """opFeatures(obj) ... return all standard features and edges based geometries"""
        return (
            PathOp.FeatureTool
            | PathOp.FeatureHeights
            | PathOp.FeatureDepths
            | PathOp.FeatureStepDown
            | PathOp.FeatureBaseFaces
            | PathOp.FeatureCoolant
        )

    def setupAdditionalProperties(self, obj):
        if not hasattr(obj, "BaseShapes"):
            obj.addProperty(
                "App::PropertyLinkList",
                "BaseShapes",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Additional base objects to be engraved"),
            )
        obj.setEditorMode("BaseShapes", 2)  # hide

        if not hasattr(obj, "OptimizeMovements"):

            obj.addProperty(
                "App::PropertyBool",
                "OptimizeMovements",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Optimize movements"),
            )
            obj.OptimizeMovements = False

        if not hasattr(obj, "FinishingPass"):
            obj.addProperty(
                "App::PropertyBool",
                "FinishingPass",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Add finishing pass"),
            )
            obj.FinishingPass = False

        if not hasattr(obj, "FinishingPassZOffset"):
            obj.addProperty(
                "App::PropertyDistance",
                "FinishingPassZOffset",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Finishing pass Z offset"),
            )

            obj.FinishingPassZOffset = "0.00"

    def initOperation(self, obj):
        """initOperation(obj) ... create vcarve specific properties."""
        obj.addProperty(
            "App::PropertyFloat",
            "Discretize",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The deflection value for discretizing arcs"),
        )
        obj.addProperty(
            "App::PropertyFloat",
            "Colinear",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Cutoff for removing colinear segments (degrees). \
                        default=10.0.",
            ),
        )
        obj.addProperty(
            "App::PropertyFloat",
            "Tolerance",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Vcarve Tolerance"),
        )

        obj.Colinear = 10.0
        obj.Discretize = 0.25
        obj.Tolerance = Path.Preferences.defaultGeometryTolerance()
        self.setupAdditionalProperties(obj)

    def opOnDocumentRestored(self, obj):
        # upgrade ...
        self.setupAdditionalProperties(obj)

    def buildMedialWires(self, obj, faces):
        """
        constructs a medial axis path using openvoronoi
        :returns: dictionary - each face object is a key containing list of wires"""

        wires_by_face = dict()
        self.voronoiDebugCache = dict()

        def is_exterior(vertex, face):
            vector = FreeCAD.Vector(vertex.toPoint(face.BoundBox.ZMin))
            (u, v) = face.Surface.parameter(vector)
            # isPartOfDomain is faster than face.IsInside(...)
            return not face.isPartOfDomain(u, v)

        def insert_many_wires(vd, wires):
            for wire in wires:
                Path.Log.debug("discretize value: {}".format(obj.Discretize))
                pts = wire.discretize(QuasiDeflection=obj.Discretize)
                ptv = [FreeCAD.Vector(p.x, p.y) for p in pts]
                # Check over the last point before just closing the polygon
                # by adding the start again.  If the discretizer was aiming
                # for the last point and missed by a little bit, closing the
                # polygon as is could result in OpenVoronoi truncating the
                # coordinates to a self-intersecting polygon which is invalid.
                # Instead, if the last point is close to the first, remove it
                # and let the final append close the polygon.
                # See issue 8064
                if len(ptv) > 0:
                    dist = ptv[-1].distanceToPoint(ptv[0])
                    if dist < FreeCAD.Base.Precision.confusion():
                        Path.Log.debug(
                            "Removing bad carve point: {} from polygon origin".format(dist)
                        )
                        del ptv[-1]
                ptv.append(ptv[0])

                for i in range(len(ptv) - 1):
                    vd.addSegment(ptv[i], ptv[i + 1])

        for f in faces:
            voronoiWires = []
            vd = Path.Voronoi.Diagram()
            insert_many_wires(vd, f.Wires)

            vd.construct()

            for e in vd.Edges:
                if e.isPrimary():
                    if e.isBorderline():
                        e.Color = BORDERLINE
                    else:
                        e.Color = PRIMARY
                else:
                    e.Color = SECONDARY

            # filter our colinear edged so there are fewer ones
            # to iterate over in colorExterior which is slow
            vd.colorColinear(COLINEAR, obj.Colinear)

            vd.colorExterior(EXTERIOR1)
            vd.colorExterior(EXTERIOR2, lambda v: is_exterior(v, f))

            # if colorTwin is done before colorExterior we seem to have
            # much more weird exterior edges needed to be filtered out,
            # keep it here to be safe
            vd.colorTwins(TWIN)

            wires = _collectVoronoiWires(vd)
            wires = _sortVoronoiWires(wires)
            voronoiWires.extend(wires)

            wires_by_face[f] = voronoiWires
            self.voronoiDebugCache = wires_by_face

        return wires_by_face

    def buildCommandList(self, obj, faces):
        """
        Build command list to cut wires - based on voronoi
        wire list from buildMedialWires
        """

        def getCurrentPosition(wire):
            """
            Calculate CNC head position assuming it reached the end of the wire
            """

            if not wire:
                return None

            lastEdge = wire[-1]
            return lastEdge.valueAt(lastEdge.LastParameter)

        def cutWires(wires, pathlist, optimizeMovements=False):
            currentPosition = None
            for w in wires:
                pWire = _getPartEdges(obj, w, geom)
                if pWire:
                    pathlist.extend(_cutWire(pWire, currentPosition))

                    # movement optimization only works if we provide current head position
                    if optimizeMovements:
                        currentPosition = getCurrentPosition(pWire)

        def canSkipRepositioning(currentPosition, newPosition):
            """
            Calculate if it makes sense to raise head to safe height and reposition before
            starting to cut another edge
            """

            if not currentPosition:
                return False

            # get vertex position on X/Y plane only
            v0 = FreeCAD.Base.Vector(currentPosition.x, currentPosition.y)
            v1 = FreeCAD.Base.Vector(newPosition.x, newPosition.y)

            return v0.distanceToPoint(v1) <= 0.5

        def _cutWire(wire, currentPosition=None):
            path = []

            e = wire[0]
            newPosition = e.valueAt(e.FirstParameter)

            # raise and reposition the head only if new wire starts further than 0.5 mm
            # from current head position
            if not canSkipRepositioning(currentPosition, newPosition):
                path.append(Path.Command("G0", {"Z": obj.SafeHeight.Value}))
                path.append(
                    Path.Command(
                        "G0", {"X": newPosition.x, "Y": newPosition.y, "Z": obj.SafeHeight.Value}
                    )
                )

            hSpeed = obj.ToolController.HorizFeed.Value
            vSpeed = obj.ToolController.VertFeed.Value
            path.append(
                Path.Command(
                    "G1", {"X": newPosition.x, "Y": newPosition.y, "Z": newPosition.z, "F": vSpeed}
                )
            )
            for e in wire:
                path.extend(Path.Geom.cmdsForEdge(e, hSpeed=hSpeed, vSpeed=vSpeed))

            return path

        pathlist = []
        pathlist.append(Path.Command("(starting)"))

        geom = _Geometry.FromObj(obj, self.model[0])

        # iterate over each face separately
        for face, wires in self.buildMedialWires(obj, faces).items():

            # If using depth step-down, calculate maximum usable depth for current face.
            # This is done to avoid adding additional step-down engraving passes when it
            # would make no sense as depth is limited by Maximum Inscribed Circle anyway.

            maximumUsableDepth = geom.stop

            if geom.stepDown > 0:
                _maximumUsableDepth = _get_maximumUsableDepth(wires, geom)
                if _maximumUsableDepth is not None:
                    maximumUsableDepth = _maximumUsableDepth
                    Path.Log.debug(f"Maximum usable depth for current face: {maximumUsableDepth}")

            # first pass
            cutWires(wires, pathlist, obj.OptimizeMovements)

            # subsequent stepDown depth passes (if any)
            while geom.incrementStepDownDepth(maximumUsableDepth):
                cutWires(wires, pathlist, obj.OptimizeMovements)

            # add finishing pass if enabled

            if obj.FinishingPass:
                geom.offset = obj.FinishingPassZOffset.Value

                cutWires(wires, pathlist, obj.OptimizeMovements)

        self.commandlist = pathlist

    def opExecute(self, obj):
        """opExecute(obj) ... process engraving operation"""
        Path.Log.track()

        self.voronoiDebugCache = None

        if obj.ToolController is None:
            return

        if not hasattr(obj.ToolController.Tool, "CuttingEdgeAngle"):
            Path.Log.info(
                translate(
                    "CAM_Vcarve",
                    "VCarve requires an engraving cutter with a cutting edge angle",
                )
            )
            return

        if obj.ToolController.Tool.CuttingEdgeAngle >= 180.0:
            Path.Log.info(
                translate("CAM_Vcarve", "Engraver cutting edge angle must be < 180 degrees.")
            )
            return

        try:
            faces = []

            for base in obj.BaseShapes:
                faces.extend(base.Shape.Faces)

            for base in obj.Base:
                for sub in base[1]:
                    shape = getattr(base[0].Shape, sub)
                    if isinstance(shape, Part.Face):
                        faces.append(shape)

            if not faces:
                for model in self.model:
                    if model.isDerivedFrom("Sketcher::SketchObject") or model.isDerivedFrom(
                        "Part::Part2DObject"
                    ):
                        faces.extend(model.Shape.Faces)

            if faces:
                self.buildCommandList(obj, faces)
            else:
                Path.Log.error(
                    translate(
                        "PathVcarve",
                        "The Job Base Object has no engraveable element. Engraving operation will produce no output.",
                    )
                )

        except Exception:
            Path.Log.warning(
                "Error processing Base object. Engraving operation will produce no output."
            )
            import traceback

            Path.Log.error(f"Engraving operation exception: {traceback.format_exc()}")

    def opUpdateDepths(self, obj, ignoreErrors=False):
        """updateDepths(obj) ... engraving is always done at the top most z-value"""
        job = PathUtils.findParentJob(obj)
        self.opSetDefaultValues(obj, job)

    def opSetDefaultValues(self, obj, job):
        """opSetDefaultValues(obj) ... set depths for vcarving"""
        if PathOp.FeatureDepths & self.opFeatures(obj):
            if job and len(job.Model.Group) > 0:
                bb = job.Proxy.modelBoundBox(job)
                obj.OpStartDepth = bb.ZMax
                obj.OpFinalDepth = job.Stock.Shape.BoundBox.ZMin
            else:
                obj.OpFinalDepth = -0.1

    def isToolSupported(self, obj, tool):
        """isToolSupported(obj, tool) ... returns True if v-carve op can work with tool."""
        return (
            hasattr(tool, "Diameter")
            and hasattr(tool, "CuttingEdgeAngle")
            and hasattr(tool, "TipDiameter")
        )

    def debugVoronoi(self, obj):
        """Debug function to display calculated voronoi edges"""

        if not getattr(self, "voronoiDebugCache", None):
            Path.Log.error("debugVoronoi: empty debug cache. Recompute VCarve operation first")
            return

        vPart = FreeCAD.activeDocument().addObject("App::Part", f"{obj.Name}-VoronoiDebug")

        wiresToShow = []

        for face, wires in self.voronoiDebugCache.items():
            for wire in wires:
                currentPartWire = Part.Wire()
                currentPartWire.fixTolerance(0.01)
                for edge in wire:
                    currentEdge = edge.toShape()

                    for v in currentEdge.Vertexes:
                        v.fixTolerance(0.1)

                    currentPartWire.add(currentEdge)
                wiresToShow.append(currentPartWire)

        for w in wiresToShow:
            vPart.addObject(Part.show(w))


def SetupProperties():
    return ["Discretize"]


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Vcarve operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectVcarve(obj, name, parentJob)
    return obj



# ***************************************************************************
# *   Copyright (c) 2002,2003 Juergen Riegel <juergen.riegel@web.de>        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# Test gui init module


class TestWorkbench(Workbench):
    "Test workbench object"

    def __init__(self):
        self.__class__.Icon = (
            FreeCAD.getResourceDir() + "Mod/Test/Resources/icons/TestWorkbench.svg"
        )
        self.__class__.MenuText = "Test Framework"
        self.__class__.ToolTip = "Test Framework"

    def Initialize(self):
        import TestGui

        list = ["Test_Test", "Test_TestAll", "Test_TestDoc", "Test_TestBase"]
        self.appendToolbar("TestTools", list)

        menu = ["Test &Commands", "TestToolsGui"]
        list = [
            "Std_TestQM",
            "Std_TestReloadQM",
            "Test_Test",
            "Test_TestAll",
            "Test_TestDoc",
            "Test_TestBase",
        ]
        self.appendCommandbar("TestToolsGui", list)
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "TestToolsText"]
        list = ["Test_TestAllText", "Test_TestDocText", "Test_TestBaseText"]
        self.appendCommandbar("TestToolsText", list)
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "TestToolsMenu"]
        list = ["Test_TestCreateMenu", "Test_TestDeleteMenu", "Test_TestWork"]
        self.appendCommandbar("TestToolsMenu", list)
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "TestFeatureMenu"]
        list = ["Test_InsertFeature"]
        self.appendCommandbar("TestFeature", list)
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "Progress bar"]
        list = [
            "Std_TestProgress1",
            "Std_TestProgress2",
            "Std_TestProgress3",
            "Std_TestProgress4",
            "Std_TestProgress5",
        ]
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "Console"]
        list = ["Std_TestConsoleOutput"]
        self.appendMenu(menu, list)

        menu = ["Test &Commands", "MDI"]
        list = ["Std_MDITest1", "Std_MDITest2", "Std_MDITest3"]
        self.appendMenu(menu, list)

        list = ["Std_ViewExample1", "Std_ViewExample2", "Std_ViewExample3"]
        self.appendMenu("Inventor View", list)


Gui.addWorkbench(TestWorkbench())

# Base system tests
FreeCAD.__unit_test__ += [
    "Workbench",
    "Menu",
    "Menu.MenuDeleteCases",
    "Menu.MenuCreateCases",
    "GuiDocument",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

from typing import Any, Final

from Base.BaseClass import BaseClass
from Base.Metadata import constmethod, export

@export(
    Include="Mod/CAM/App/VoronoiEdge.h",
    Namespace="Path",
    RichCompare=True,
    Constructor=True,
    Delete=True,
)
class VoronoiEdge(BaseClass):
    """
    Edge of a Voronoi diagram

    Author: sliptonic (shopinthewoods@gmail.com)
    License: LGPL-2.1-or-later
    """

    @constmethod
    def isFinite(self) -> Any:
        """Returns true if both vertices are finite"""
        ...

    @constmethod
    def isInfinite(self) -> Any:
        """Returns true if the end vertex is infinite"""
        ...

    @constmethod
    def isLinear(self) -> Any:
        """Returns true if edge is straight"""
        ...

    @constmethod
    def isCurved(self) -> Any:
        """Returns true if edge is curved"""
        ...

    @constmethod
    def isPrimary(self) -> Any:
        """Returns false if edge goes through endpoint of the segment site"""
        ...

    @constmethod
    def isSecondary(self) -> Any:
        """Returns true if edge goes through endpoint of the segment site"""
        ...

    @constmethod
    def isBorderline(self) -> Any:
        """Returns true if the point is on the segment"""
        ...

    @constmethod
    def toShape(self) -> Any:
        """Returns a shape for the edge"""
        ...

    @constmethod
    def getDistances(self) -> Any:
        """Returns the distance of the vertices to the input source"""
        ...

    @constmethod
    def getSegmentAngle(self) -> Any:
        """Returns the angle (in degree) of the segments if the edge was formed by two segments"""
        ...
    Index: Final[int]
    """Internal id of the element."""

    Color: int
    """Assigned color of the receiver."""

    Cell: Final[Any]
    """cell the edge belongs to"""

    Vertices: Final[list]
    """Begin and End voronoi vertex"""

    Next: Final[Any]
    """CCW next edge within voronoi cell"""

    Prev: Final[Any]
    """CCW previous edge within voronoi cell"""

    RotNext: Final[Any]
    """Rotated CCW next edge within voronoi cell"""

    RotPrev: Final[Any]
    """Rotated CCW previous edge within voronoi cell"""

    Twin: Final[Any]
    """Twin edge"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# **************************************************************************
#   Copyright (c) 2023 David Carter <dcarter@davidcarter.ca>              *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
# **************************************************************************

"""
Test module for FreeCAD material cards and APIs
"""
import os

import unittest
import FreeCAD
import Materials

parseQuantity = FreeCAD.Units.parseQuantity

UUIDAcrylicLegacy = ""  # This can't be known until it is loaded
UUIDAluminumAppearance = "3c6d0407-66b3-48ea-a2e8-ee843edf0311"
UUIDAluminumMixed = "5f546608-fcbb-40db-98d7-d8e104eb33ce"
UUIDAluminumPhysical = "a8e60089-550d-4370-8e7e-1734db12a3a9"
UUIDBrassAppearance = "fff3d5c8-98c3-4ee2-8fe5-7e17403c48fcc"


class MaterialFilterTestCases(unittest.TestCase):
    """
    Test class for FreeCAD material cards and APIs
    """

    def setUp(self):
        """Setup function to initialize test data"""
        self.ModelManager = Materials.ModelManager()
        self.MaterialManager = Materials.MaterialManager()
        self.uuids = Materials.UUIDs()

        # Use our test files as a custom directory
        param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/Resources")
        self.customDir = param.GetString("CustomMaterialsDir", "")
        self.useBuiltInDir = param.GetBool("UseBuiltInMaterials", True)
        self.useWorkbenchDir = param.GetBool("UseMaterialsFromWorkbenches", True)
        self.useUserDir = param.GetBool("UseMaterialsFromConfigDir", True)
        self.useCustomDir = param.GetBool("UseMaterialsFromCustomDir", False)

        paramExternal = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/ExternalInterface")
        self.useExternal = paramExternal.GetBool("UseExternal", False)

        filePath = os.path.dirname(__file__) + os.sep
        testPath = filePath + "Materials"
        param.SetString("CustomMaterialsDir", testPath)
        param.SetBool("UseBuiltInMaterials", False)
        param.SetBool("UseMaterialsFromWorkbenches", False)
        param.SetBool("UseMaterialsFromConfigDir", False)
        param.SetBool("UseMaterialsFromCustomDir", True)

        paramExternal.SetBool("UseExternal", False)

        self.MaterialManager.refresh()

    def tearDown(self):
        param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/Resources")

        # Restore preferences
        param.SetString("CustomMaterialsDir", self.customDir)
        param.SetBool("UseBuiltInMaterials", self.useBuiltInDir)
        param.SetBool("UseMaterialsFromWorkbenches", self.useWorkbenchDir)
        param.SetBool("UseMaterialsFromConfigDir", self.useUserDir)
        param.SetBool("UseMaterialsFromCustomDir", self.useCustomDir)

        paramExternal = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/ExternalInterface")
        paramExternal.SetBool("UseExternal", self.useExternal)

        self.MaterialManager.refresh()

    def testFilter(self):
        """Test that our filter returns the correct materials"""

        # First check that our materials are loading
        material = self.MaterialManager.getMaterial(UUIDAluminumAppearance)
        self.assertIsNotNone(material)
        self.assertEqual(material.Name, "TestAluminumAppearance")
        self.assertEqual(material.UUID, UUIDAluminumAppearance)

        material = self.MaterialManager.getMaterial(UUIDAluminumMixed)
        self.assertIsNotNone(material)
        self.assertEqual(material.Name, "TestAluminumMixed")
        self.assertEqual(material.UUID, UUIDAluminumMixed)

        material = self.MaterialManager.getMaterial(UUIDAluminumPhysical)
        self.assertIsNotNone(material)
        self.assertEqual(material.Name, "TestAluminumPhysical")
        self.assertEqual(material.UUID, UUIDAluminumPhysical)

        material = self.MaterialManager.getMaterial(UUIDBrassAppearance)
        self.assertIsNotNone(material)
        self.assertEqual(material.Name, "TestBrassAppearance")
        self.assertEqual(material.UUID, UUIDBrassAppearance)

        # Create an empty filter
        filter = Materials.MaterialFilter()
        self.assertEqual(len(self.MaterialManager.MaterialLibraries), 1)

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 4)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 5)

        # Create a basic rendering filter
        filter.Name = "Basic Appearance"
        filter.RequiredCompleteModels = [self.uuids.BasicRendering]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 3)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 3)

        # Create an advanced rendering filter
        filter= Materials.MaterialFilter()
        filter.Name = "Advanced Appearance"
        filter.RequiredCompleteModels = [self.uuids.AdvancedRendering]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 0)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 0)

        # Create a Density filter
        filter= Materials.MaterialFilter()
        filter.Name = "Density"
        filter.RequiredCompleteModels = [self.uuids.Density]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 2)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 3)

        # Create a Hardness filter
        filter= Materials.MaterialFilter()
        filter.Name = "Hardness"
        filter.RequiredCompleteModels = [self.uuids.Hardness]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 0)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 0)

        # Create a Density and Basic Rendering filter
        filter= Materials.MaterialFilter()
        filter.Name = "Density and Basic Rendering"
        filter.RequiredCompleteModels = [self.uuids.Density, self.uuids.BasicRendering]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 1)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 1)

        # Create a Linear Elastic filter
        filter= Materials.MaterialFilter()
        filter.Name = "Linear Elastic"
        filter.RequiredCompleteModels = [self.uuids.LinearElastic]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 0)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)
        self.assertEqual(len(filtered), 0)

        filter= Materials.MaterialFilter()
        filter.Name = "Linear Elastic - incomplete"
        filter.RequiredModels = [self.uuids.LinearElastic]

        filtered = self.MaterialManager.filterMaterials(filter)
        self.assertEqual(len(filtered), 2)

        filtered = self.MaterialManager.filterMaterials(filter, includeLegacy=True)

    def testErrorInput(self):

        self.assertRaises(TypeError, self.MaterialManager.filterMaterials, [])



from typing import Any, Final

from Base.PyObjectBase import PyObjectBase
from Base.Metadata import constmethod, export

@export(
    Include="Mod/TechDraw/App/Cosmetic.h",
    Namespace="TechDraw",
    Constructor=True,
    Delete=True,
)
class CosmeticVertex(PyObjectBase):
    """
    CosmeticVertex specifies an extra (cosmetic) vertex in Views
    
    Author: WandererFan (wandererfan@gmail.com)
    License: LGPL-2.1-or-later
    """

    @constmethod
    def clone(self) -> Any:
        """Create a clone of this CosmeticVertex"""
        ...

    @constmethod
    def copy(self) -> Any:
        """Create a copy of this CosmeticVertex"""
        ...
    Tag: Final[str]
    """Gives the tag of the CosmeticVertex as string."""

    Point: Any
    """Gives the position of this CosmeticVertex as vector."""

    Show: bool
    """Show/hide the vertex."""

    Color: Any # type: tuple[float, float, float, float]]
    """set/return the vertex's colour using a tuple (rgba)."""

    Size: Any
    """set/return the vertex's radius in mm."""

    Style: Any
    """set/return the vertex's style as integer."""



#!/usr/bin/env python3
import argparse
import sys
import os
import re
import logging
from utils import (
    run_command,
    init_environment,
    add_common_arguments,
    emit_problem_matchers,
    write_file,
    append_file,
)


def generate_markdown_report(
    errors_count: int,
    warnings_count: int,
    notes_count: int,
    clang_tidy_output: str,
    enabled_checks_content: str,
) -> str:
    """Generate a Markdown report section for clang-tidy results and enabled checks."""
    report_lines = []
    if errors_count > 0:
        report_lines.append(
            f"<details><summary>:fire: Clang-Tidy found :fire: {errors_count} errors, "
            f":warning: {warnings_count} warnings and :pencil2: {notes_count} notes</summary>"
        )
    elif warnings_count > 0:
        report_lines.append(
            f"<details><summary>:warning: Clang-Tidy found :warning: {warnings_count} "
            f"warnings and :pencil2: {notes_count} notes</summary>"
        )
    elif notes_count > 0:
        report_lines.append(
            f"<details><summary>:pencil2: Clang-Tidy found :pencil2: {notes_count} notes</summary>"
        )
    else:
        report_lines.append(
            "<details><summary>:heavy_check_mark: Clang-Tidy found no errors, warnings or notes</summary>"
        )
    report_lines.append("")
    report_lines.append("````")
    report_lines.append(clang_tidy_output)
    report_lines.append("````")
    report_lines.append("</details>")
    report_lines.append("")
    report_lines.append(
        "<details><summary>:information_source: Enabled checks</summary>"
    )
    report_lines.append("")
    report_lines.append("````")
    report_lines.append(enabled_checks_content)
    report_lines.append("````")
    report_lines.append("</details>")
    report_lines.append("")
    return "\n".join(report_lines)


def count_occurrences(pattern: str, text: str) -> int:
    """Count all occurrences of a regex pattern in text using MULTILINE mode."""
    matches = re.findall(pattern, text, re.MULTILINE)
    logging.info("Pattern '%s' found %d matches", pattern, len(matches))
    return len(matches)


def main():
    parser = argparse.ArgumentParser(
        description="Run clang-tidy on provided C++ files and append a Markdown report."
    )
    add_common_arguments(parser)
    parser.add_argument(
        "--clang-style",
        required=True,
        help="Clang-format style (e.g., 'file' to use .clang-format or a specific style).",
    )
    parser.add_argument(
        "--line-filter",
        required=False,
        help='Line-filter for clang-tidy (i.e. [{"name":"file1.cpp","lines":[[1,3],[5,7]]},...])',
    )
    args = parser.parse_args()
    init_environment(args)

    build_dir = "build"

    clang_tidy_base_cmd = [
        "clang-tidy",
        "--quiet",
        f"--format-style={args.clang_style}",
        "--export-fixes=clang-tidy.yaml",
        "-p",
        build_dir,
    ]

    enabled_cmd = clang_tidy_base_cmd + ["--explain-config"]
    enabled_stdout, enabled_stderr, _ = run_command(enabled_cmd)
    enabled_output = enabled_stdout + enabled_stderr
    enabled_checks_log = os.path.join(args.log_dir, "clang-tidy-enabled-checks.log")
    write_file(enabled_checks_log, enabled_output)

    clang_cmd = clang_tidy_base_cmd
    if args.line_filter:
        clang_cmd = clang_cmd + [f"--line-filter={args.line_filter}"]
    clang_cmd = clang_cmd + args.files.split()
    print("clang_cmd = ", clang_cmd)
    clang_stdout, clang_stderr, _ = run_command(clang_cmd)
    clang_tidy_output = clang_stdout + clang_stderr

    clang_tidy_log = os.path.join(args.log_dir, "clang-tidy.log")
    write_file(clang_tidy_log, clang_tidy_output)
    emit_problem_matchers(clang_tidy_log, "clang.json", "clang")

    error_pattern = r"^.+:\d+:\d+: error: .+$"
    warning_pattern = r"^.+:\d+:\d+: warning: .+$"
    note_pattern = r"^.+:\d+:\d+: note: .+$"

    errors_count = count_occurrences(error_pattern, clang_tidy_output)
    warnings_count = count_occurrences(warning_pattern, clang_tidy_output)
    notes_count = count_occurrences(note_pattern, clang_tidy_output)

    logging.info(
        "Found %d errors, %d warnings, %d notes",
        errors_count,
        warnings_count,
        notes_count,
    )

    report = generate_markdown_report(
        errors_count, warnings_count, notes_count, clang_tidy_output, enabled_output
    )
    append_file(args.report_file, report)

    sys.exit(0 if errors_count == 0 else 1)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

from typing import Any, Final

from Base.Metadata import constmethod, export
from Base.Persistence import Persistence

@export(
    Include="Mod/CAM/App/Path.h",
    Twin="Toolpath",
    TwinPointer="Toolpath",
    Namespace="Path",
    Delete=True,
    Constructor=True,
)
class Path(Persistence):
    """
    Path([commands]): Represents a basic Gcode path
    commands (optional) is a list of Path commands

    Author: Yorik van Havre (yorik@uncreated.net)
    License: LGPL-2.1-or-later
    """

    def addCommands(self) -> Any:
        """adds a command or a list of commands at the end of the path"""
        ...

    def insertCommand(self) -> Any:
        """insertCommand(Command,[int]):
        adds a command at the given position or at the end of the path"""
        ...

    def deleteCommand(self) -> Any:
        """deleteCommand([int]):
        deletes the command found at the given position or from the end of the path"""
        ...

    def setFromGCode(self) -> Any:
        """sets the contents of the path from a gcode string"""
        ...

    @constmethod
    def toGCode(self) -> Any:
        """returns a gcode string representing the path"""
        ...

    @constmethod
    def copy(self) -> Any:
        """returns a copy of this path"""
        ...

    @constmethod
    def getCycleTime(self) -> Any:
        """return the cycle time estimation for this path in s"""
        ...
    Length: Final[float]
    """the total length of this path in mm"""

    Size: Final[int]
    """the number of commands in this path"""

    Commands: list
    """the list of commands of this path"""

    Center: Any
    """the center position for all rotational parameters"""

    BoundBox: Final[Any]
    """the extent of this path"""



from Base.Metadata import export
from DrawView import DrawView

@export(
    Include="Mod/TechDraw/App/DrawLeaderLine.h",
    Namespace="TechDraw",
    FatherInclude="Mod/TechDraw/App/DrawViewPy.h",
)
class DrawLeaderLine(DrawView):
    """
    Feature for adding leaders to Technical Drawings
    
    Author: WandererFan (wandererfan@gmail.com)
    License: LGPL-2.1-or-later
    """
